@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2019-2022 -- leonerd@leonerd.org.uk
 
-package Object::Pad 0.68;
+package Object::Pad 0.69;
 
 use v5.14;
 use warnings;
@@ -1054,8 +1054,18 @@ sub _import_configuration
       my $sym = $syms->[$i];
 
       if( $sym =~ m/^:config\((.*)\)$/ ) {
-         my $opts = $1 =~ s/^\s+|\s+$//gr; # trim
-         $^H{"Object::Pad/configure($_)"}++ for split m/\s+/, $opts;
+         foreach my $opt ( split m/\s+/, $1 =~ s/^\s+|\s+$//gr ) {
+            if( $opt =~ m/^only_class_attrs=(.*)$/ ) {
+               # Store an entire sub-hash inside the hints hash. This won't
+               # survive squashing into a COP for runtime but we only need it
+               # during compile so that's OK
+               my $attrs = $1;
+               $^H{"Object::Pad/configure(only_class_attrs)"} = { map { $_ => 1 } split m/,/, $attrs };
+            }
+            else {
+               $^H{"Object::Pad/configure($opt)"}++
+            }
+         }
       }
       else {
          $i++;
